import numpy as np

from ..filters.spatial import multilook


def coherence(in1, in2, window_size=7, *args, **kwargs):
    """[summary]

    Calc coherence between two complex arrays

    Parameters
    ----------
    in1 : [type]
        [description]
    in2 : [type]
        [description]
    window_size : int, optional
        [description], by default 7
    """
    coh = multilook(in1 * np.conj(in2), window_size, *args, **kwargs) / (
        np.sqrt(
            multilook(in1 * np.conj(in1), window_size, *args, **kwargs)
            * multilook(in2 * np.conj(in2, window_size, *args, **kwargs))
        )
    )

    return np.clip(np.nan_to_num(coh), 0.0, 1.0)



import numpy as np
from scipy.ndimage import filters


def multilook(inputs, window_size=3, output=None, *args, **kwargs):
    """Multi-look or a spatial averaging over an array

    Multidimensional mean filter over an array.

    Parameters
    ----------
    inputs : array_like
        [description]
    window_size : int, optional
        N-look over x-axis(Azimuth) and y-axis (Range) , by default 3 (Azimuth=3, Range=3)
    output : array or dtype, optional
        The array in which to place the output, or the dtype of the returned array. By default an array of the same dtype as input will be created. , by default None

    Returns
    -------
    mean_filter : ndarray
        Filtered array. Has the same shape as input.

    """

    # Window size specified as a single odd integer (3, 5, 7, …), or an iterable of length image.ndim containing only odd integers (e.g. (1, 5, 5)).
    # multi-look processing or spatial averaging

    inputs = np.asarray(inputs)

    window_size = np.asarray(window_size)
    if window_size.shape == ():
        window_size = np.repeat(window_size.item(), inputs.ndim)

    if np.iscomplexobj(inputs):
        return filters.uniform_filter(
            inputs.real, window_size, output=output, *args, **kwargs
        ) + 1j * filters.uniform_filter(
            inputs.imag, window_size, output=output, *args, **kwargs
        )
    else:
        return filters.uniform_filter(
            inputs.real, window_size, output=output, *args, **kwargs
        )



import numpy as np


def ent_ani_alp(ew, ev):
    p = pseudo_probabilities(ew)
    entropy_h = entropy(p)
    _, alpha_mean = alpha_angles(ev, p)
    anisotropy_a = anisotropy(p)
    print(
        "Eigenvalues shape: {0}\t Eigenvectors shape: {1}".format(
            ew.shape, ev.shape
        )
    )
    return np.stack([entropy_h, anisotropy_a, alpha_mean])


def pseudo_probabilities(ew):
    p = np.zeros(ew.shape, dtype=float)
    p = ew[:, :, :] / np.sum(ew, axis=2)[:, :, np.newaxis]

    assert np.allclose(
        np.sum(p[:, :, :], axis=2), 1
    ), "Pseudo Probabilities should be 1 along axis 2"
    return p


def entropy(p):
    entropy_h = -np.sum(p * np.log(p) / np.log(p.shape[2]), axis=2)
    assert (
        np.max(entropy_h) <= 1 and np.min(entropy_h) >= 0
    ), "Entopry should be in range 0 < H < 1"
    return entropy_h


def anisotropy(p):
    anisotropy_a = (p[:, :, -2] - p[:, :, -3]) / (
        p[:, :, -2] + p[:, :, -3]
    )  # (lamda1 - lamda 2) / (lamda1 + lamda2)
    return anisotropy_a


def alpha_angles(ev, p):
    alpha = np.zeros_like(p, dtype=np.float)
    # print(ev.shape, p.shape)
    assert np.allclose(np.sum(ev[:, :, 0, :].imag, axis=2), 0), (
        "Eigenvector" "s first element imag part should be almost zero."
    )
    alpha[:, :, :] = np.arccos(
        np.abs(ev[:, :, 0, :])
    )  # cos of each top element of each vector

    alpha_mean = np.sum(alpha * p, axis=2)

    return alpha, alpha_mean



import datetime

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.colors import LogNorm
from matplotlib.ticker import PercentFormatter

from .plot import image_dB

plt.rcParams["image.origin"] = "lower"

# Save a palette to a variable:\

colors = ["#003FFF", "#E8000B", "#03ED3A", "#1A1A1A"]  # B, R, G, Black
colors = ["#E8000B", "#03ED3A", "#003FFF", "#1A1A1A"]  # R, G, B,Black
colors = ["#003FFF", "#E8000B", "#03ED3A", "#1A1A1A"]  # Blue, Red, Green, Black

palette = sns.color_palette(colors, 4)


def plot_t_c_matrix(
    c_t_matrix,
    title="T",
    vmin=2,
    vmax=98,
    aspect=None,
    origin="lower",
    *args,
    **kwargs,
):
    C_T_11 = c_t_matrix[:, :, 0, 0]
    im_perc = np.percentile(np.abs(C_T_11), [vmin, vmax])
    if aspect is None:
        aspect = c_t_matrix.shape[1] / c_t_matrix.shape[0] * 2

    def plot_sections(c_t_matrix):

        fig, axs = plt.subplots(3, 4, constrained_layout=True)
        for idx, ax in zip(range(len(c_t_matrix)), axs[0, :3].flat):
            # print(idx)
            c_idx_abs = np.abs(c_t_matrix[:, :, idx, idx])
            pcm = ax.imshow(
                c_idx_abs,
                interpolation="hamming",
                norm=LogNorm(
                    vmin=im_perc[0], vmax=im_perc[1]
                ),  # vmin=.9*c11_db.min() , vmax= 1/3*c11_db.max(),
                cmap="gray",
                aspect=aspect,
                origin=origin,
                *args,
                **kwargs,
            )
            ax.axis("off")
            ax.axis("off")
            ax.set_title("${0}_{{{1}{1}}}$".format(title, idx + 1))

            # plot hisogram
            ax_hist = axs[0, -1]
            ax_hist.hist(
                image_dB(c_idx_abs.ravel()),
                bins="auto",
                color=palette[idx],
                histtype="step",
                label="${0}_{{{1}{1}}}$".format(title, idx + 1),
                density=True,
                linestyle=("solid"),
            )
            ax_hist.legend(loc="upper left")  # , bbox_to_anchor=(0.5, 1.05),
            #  ncol=3, fancybox=True, shadow=True)
            ax_hist.set_title("Histogram")

        cbar_1 = fig.colorbar(pcm, ax=axs[0, :3], shrink=0.6, location="left")
        cbar_1.set_label("$dB $ ", rotation=0)

        # plot of images
        il_lower = np.tril_indices(c_t_matrix.shape[3], -1)
        for idx, nx, ny, ax in zip(
            range(4), il_lower[0], il_lower[1], axs[1, :3].flat
        ):
            #
            # print(nx, ny)
            c_idx_abs = np.abs(c_t_matrix[:, :, nx, ny])
            c_idx_angle = np.angle(c_t_matrix[:, :, nx, ny])

            pcm = ax.imshow(
                c_idx_abs,
                interpolation="hamming",
                norm=LogNorm(
                    vmin=im_perc[0], vmax=im_perc[1]
                ),  # vmin=.9*c11_db.min() , vmax= 1/3*c11_db.max(),
                cmap="gray",
                aspect=aspect,
                origin=origin,
                *args,
                **kwargs,
            )
            ax.axis("off")
            ax.axis("off")
            ax.set_title("${0}_{{{1}{2}}}$".format(title, nx + 1, ny + 1))

            # Histogram
            ax_hist = axs[1, -1]
            ax_hist.hist(
                image_dB(c_idx_abs.ravel()),
                bins="auto",
                color=palette[idx],
                histtype="step",
                label="${0}_{{{1}{2}}}$".format(title, nx + 1, ny + 1),
                density=True,
                linestyle=("solid"),
            )

            ax_hist.legend(loc="upper left")  # , bbox_to_anchor=(0.5, 1.05),
            #  ncol=3, fancybox=True, shadow=True)
            ax_hist.set_title("Histogram")

            ax_phase = axs[2, idx]
            pcm_phase = ax_phase.imshow(
                c_idx_angle,
                interpolation="hamming",
                vmin=-np.pi,
                vmax=np.pi,  # vmin=.9*c11_db.min() , vmax= 1/3*c11_db.max(),
                cmap="jet",
                aspect=aspect,
                origin=origin,
                *args,
                **kwargs,
            )
            ax_phase.axis("off")
            ax_phase.axis("off")
            ax_phase.set_title(
                "$\phi_{{ {0}_{{{1}{2}}} }}$".format(title, nx + 1, ny + 1)
            )

            # Histogram
            ax_phase_hist = axs[2, -1]
            ax_phase_hist.hist(
                c_idx_angle.ravel(),
                bins="auto",
                color=palette[idx],
                histtype="step",
                label="$\phi_{{ {0}_{{{1}{2}}} }}$".format(
                    title, nx + 1, ny + 1
                ),
                density=True,
                linestyle=("solid"),
            )

            ax_phase_hist.legend(
                loc="upper left"
            )  # , bbox_to_anchor=(0.5, 1.05),
            #  ncol=3, fancybox=True, shadow=True)
            ax_phase_hist.set_title("Histogram")

        ax_phase_hist.set_xlim([-np.pi, np.pi])
        ax_phase_hist.set_xticks([-np.pi, 0, np.pi])
        ax_phase_hist.set_xticklabels(["$- \pi$", "0", "$\pi$"])

        cbar_2 = fig.colorbar(pcm, ax=axs[1, :3], shrink=0.6, location="left")
        cbar_2.set_label("$dB $ ", rotation=0)
        cbar_3 = fig.colorbar(
            pcm_phase,
            ax=axs[2, :3],
            shrink=0.6,
            location="left",
            ticks=[-np.pi, 0, np.pi],
        )
        cbar_3.ax.set_yticklabels(["$- \pi$", "0", "$\pi$"])
        cbar_3.set_label("$\phi$", rotation=0)
        return fig

    fig = plot_sections(c_t_matrix)
    return fig


def plot_eigenvalues_hist_dB(
    array_dict,
    suptitle,
    vmin=2,
    vmax=98,
    aspect=None,
    origin="lower",
    *args,
    **kwargs,
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)
    # sns.palplot(palette)

    n_images = len(array_dict.keys())
    fig = plt.figure(constrained_layout=False, figsize=(4 * n_images, 10))
    gs = fig.add_gridspec(nrows=2, ncols=n_images)
    f_img = {}
    # f_hist = {}
    ax_img = {}
    f_hist = fig.add_subplot(gs[1, :])

    for idx, key in enumerate(array_dict):
        legend = key
        array = array_dict[key]
        if aspect is None:
            aspect = array.shape[1] / array.shape[0] * 2

        f_img[idx] = fig.add_subplot(gs[0, idx])
        im_perc = np.percentile(array, [vmin, vmax])

        ax_img[idx] = f_img[idx].imshow(
            array,
            aspect=aspect,
            cmap="jet",
            vmin=0,
            vmax=im_perc[1],
            origin=origin,
        )
        f_img[idx].axis("off")
        f_img[idx].set_title(legend)
        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_img[idx],
            ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=0.8,
            pad=0.1,
        )
        cbar.ax.set_xlabel("", rotation=0)

        f_hist.hist(
            image_dB(array.flatten()),
            color=palette[idx],
            bins="auto",
            histtype="step",
            label=legend,
            density=True,
        )  # , linestyle=(':')
        f_hist.legend()
        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))
    f_hist.set_xlabel("$dB$")
    f_hist.set_ylabel("Normalized Frequency")
    fig.suptitle(suptitle)
    # now = datetime.datetime.now()
    # fig.savefig("eigenvalues_{}_{}.tiff".format(title, now.strftime("%Y-%m-%d-%H:%M:%S")))
    return fig


def plot_image_hist(
    array_dict, suptitle, aspect=None, origin="lower", *args, **kwargs
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)

    n_images = len(array_dict.keys())
    fig = plt.figure(constrained_layout=False, figsize=(9 * n_images, 10))
    gs = fig.add_gridspec(nrows=2, ncols=n_images + 1)

    for idx, key in enumerate(array_dict):
        title = key
        array = array_dict[key]
        ax_im = fig.add_subplot(gs[0, idx])
        if (np.max(array) - np.min(array)) < 1.1:
            ax_cc = ax_im.imshow(array, cmap="jet", vmin=0.0, vmax=1.0)
        elif (np.max(array) - np.min(array)) < 100 and (
            np.max(array) - np.min(array)
        ) > 10:
            ax_cc = ax_im.imshow(array, cmap="jet", vmin=0.0, vmax=90.0)
        else:
            ax_cc = ax_im.imshow(
                array, cmap="jet", vmin=np.min(array), vmax=np.max(array)
            )

        ax_im.axis("off")
        ax_im.set_title(title)
        # ax_im.clim(0, 10)
        ax_cc.set_clim(ax_cc.get_clim()[0], ax_cc.get_clim()[1])

        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_cc,
            ax=ax_im,  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=1.0,
            pad=0.05,
        )
        cbar.ax.set_xlabel("", rotation=0)

        ax_hist = fig.add_subplot(gs[1, idx])
        ax_hist.hist(
            (array.ravel()),
            color=palette[idx],
            bins=1000,
            histtype="step",
            label=title,
            density=False,
            weights=100 * np.ones(len(array.ravel())) / len(array.ravel()),
        )  # , linestyle=(':')
        ax_hist.set_xlim(ax_cc.get_clim())
        # Now we format the y-axis to display percentage
        ax_hist.yaxis.set_major_formatter(PercentFormatter(xmax=1))
        fig.suptitle(suptitle)

        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))

    return fig



"""

Hajnsek I., Papathanassiou K.P. & Cloude S.R.
(2001), “Removal of Additive Noise in Polarimetric
Eigenvalue Processing”, Proc. IGARSS, Sidney,
Australia, pp.2778–2780.
"""

import numpy as np


def additive_noise(c_t_matrix_44):
    """
    Inputs: T or C Matrix 4x4
    Output: T or C Matrix 3x3 (Noise Removed)
    """
    from ..decomposition.eigen import eigen_decomposition

    assert c_t_matrix_44.shape[-1] == 4
    w_4, v_4 = eigen_decomposition(c_t_matrix_44)
    noise = w_4[:, :, 0]
    noise[noise < 0] = 0
    c_t_matrix_33 = np.zeros_like(c_t_matrix_44[:, :, :3, :3])
    c_t_matrix_33 = c_t_matrix_44[:, :, :3, :3]
    # c_t_matrix_33_diagonal = c_t_matrix_33.diagonal( axis1=-2, axis2=-1) - np.repeat(noise[:,:, np.newaxis], 3, axis=2)

    """
    c_t_matrix_33[:, :, 0, 0] = c_t_matrix_33[:, :, 0, 0] - c_t_matrix_33_diagonal[:, :, 0]
    c_t_matrix_33[:, :, 1, 1] = c_t_matrix_33[:, :, 1, 1] - c_t_matrix_33_diagonal[:, :, 1]
    c_t_matrix_33[:, :, 2, 2] = c_t_matrix_33[:, :, 2, 2] - c_t_matrix_33_diagonal[:, :, 2]
    """
    print("Additive T - 2*Noise")
    c_t_matrix_33[:, :, 0, 0] = c_t_matrix_33[:, :, 0, 0] - 2 * noise
    c_t_matrix_33[:, :, 1, 1] = c_t_matrix_33[:, :, 1, 1] - 2 * noise
    c_t_matrix_33[:, :, 2, 2] = c_t_matrix_33[:, :, 2, 2] - 2 * noise

    return c_t_matrix_33



__all__ = ["eigen_decomposition"]

import numpy as np
from numba import jit, njit, prange


def eigen_decomposition(t_matrix):
    """
    eW -> Eigenvalues w0 < w1 < w2 ... < wn
    eV -> Eigenvectors
    """
    w = np.zeros(
        (t_matrix.shape[0], t_matrix.shape[1], t_matrix.shape[2]), dtype=float
    )
    v = np.zeros(
        (
            t_matrix.shape[0],
            t_matrix.shape[1],
            t_matrix.shape[2],
            t_matrix.shape[3],
        ),
        dtype=complex,
    )

    for ix, iy in np.ndindex(t_matrix.shape[0:2]):
        w[ix, iy, :], v[ix, iy, :] = np.linalg.eigh(t_matrix[ix, iy, :, :])
    w[w <= 0] = 0.0
    return w, v


"""
@jit
def eigen_decomposition_jit(t_matrix):
    """ """
    eW -> Eigenvalues w0 < w1 < w2
    eV -> Eigenvectors
    """ """
    w = np.zeros(
        (t_matrix.shape[0], t_matrix.shape[1], t_matrix.shape[2]), dtype=float
    )
    v = np.zeros(
        (
            t_matrix.shape[0],
            t_matrix.shape[1],
            t_matrix.shape[2],
            t_matrix.shape[3],
        ),
        dtype=complex,
    )

    for ix, iy in np.ndindex(t_matrix.shape[0:2]):
        w[ix, iy, :], v[ix, iy, :] = np.linalg.eigh(t_matrix[ix, iy, :, :])
    w[w <= 0] = 0.0
    return w, v
"""


@jit(nopython=True, nogil=True, cache=True, parallel=False)
def eigen_decomposition_jit(A):
    """
    eW -> Eigenvalues w0 < w1 < w2 ... < wn
    eV -> Eigenvectors
    """
    # A
    nx = A.shape[0]
    ny = A.shape[1]
    t_size = A.shape[2:]
    w = np.zeros((nx, ny, t_size[0]), dtype=np.float_)
    v = np.zeros((nx, ny, t_size[0], t_size[1],), dtype=np.complex_,)
    for i in range(nx):
        for j in range(ny):

            w[i, j, :], v[i, j, :, :] = np.linalg.eigh(
                A[i, j, :, :]
            )  # , UPLO='U' numba not supported

    return w, v


@jit(nopython=True, nogil=True, cache=True, parallel=True)
def eigen_decomposition_jit_prange(A):
    """
    eW -> Eigenvalues w0 < w1 < w2 ... < wn
    eV -> Eigenvectors
    """
    # A
    nx = A.shape[0]
    ny = A.shape[1]
    # t_size = A.shape[2:4]
    w = np.zeros((nx, ny, A.shape[2]), dtype=np.float_)
    v = np.zeros((nx, ny, A.shape[2], A.shape[3],), dtype=np.complex_,)
    for i in prange(nx):
        for j in prange(ny):

            w[i, j, :], v[i, j, :, :] = np.linalg.eigh(
                A[i, j, :, :]
            )  # , UPLO='U' numba not supported

    return w, v



def check_files_if_exist(*paths):
    from pathlib import Path

    try:
        all(Path(path).resolve(strict=True) for path in paths)
    except FileNotFoundError:
        print("File OR Folder:", "Not exist")
        raise
    else:
        # Exist
        return True


def project_quadpol_rat_files(
    DIR_campaign, ID_campaign, ID_flight, band, ID_Pass, n_try="01"
):
    """Dictionary of Full Polarimetric SAR RAT files

    Provide a dic of str for RAT files of each channel (HH, VV, HV, VH)

    Parameters
    ----------
    DIR_campaign : str
        direcotry of the campaign data
    ID_campaign : str
        ID for the campaign
    ID_flight : str
        ID for the flight
    band : str
        band of F-SAR bands=['X','C','S','L']
    ID_Pass : str
        ID pass of a given flight path
    n_try : str, optional
        Usually for Quad-Pol (n_try="01")

    Returns
    -------
    files : dict
        dict of RAT files "RAT"

    Examples
    --------
    >>> DIR_campaign = "/data/largeHome/mans_is/PolSAR-Thesis/10_users/mans_is/PolSAR-Results/01_projects/18PRMASR"
    >>> ID_campaign = '18prmasr'
    >>> ID_flight="03"
    >>> band="L"
    >>> ID_Pass="05"
    >>> n_try="01"
    >>> project_quadpol_rat_files(DIR_campaign=DIR_campaign, ID_campaign=ID_campaign, ID_flight=ID_flight, band=band, ID_Pass=ID_Pass, n_try="01")
    {'HH': '/data/HR_Projekte/Pol-InSAR_InfoRetrieval/01_projects/18PRMASR/FL03/PS05/T01L/RGI/RGI-SR/slc_18prmasr0305_Lhh_t01L.rat',
        'VV': '/data/HR_Projekte/Pol-InSAR_InfoRetrieval/01_projects/18PRMASR/FL03/PS05/T01L/RGI/RGI-SR/slc_18prmasr0305_Lvv_t01L.rat',
        'HV': '/data/HR_Projekte/Pol-InSAR_InfoRetrieval/01_projects/18PRMASR/FL03/PS05/T01L/RGI/RGI-SR/slc_18prmasr0305_Lhv_t01L.rat',
        'VH': '/data/HR_Projekte/Pol-InSAR_InfoRetrieval/01_projects/18PRMASR/FL03/PS05/T01L/RGI/RGI-SR/slc_18prmasr0305_Lvh_t01L.rat'}
    """
    from pathlib import Path

    try:
        DIR_campaign = Path(DIR_campaign).resolve(
            strict=True
        )  # .exists() # .resolve(strict=True)
        DIR_data = Path.joinpath(
            DIR_campaign,
            "FL"
            + ID_flight
            + "/"
            + "PS"
            + ID_Pass
            + "/T01"
            + band
            + "/RGI/RGI-SR/",
        )
        DIR_data = Path(DIR_data).resolve(strict=True)
    except FileNotFoundError:
        # DIR_data doesn't exist
        print("DIR_data: ", DIR_data, "Not exist")
        raise
    else:
        # exists
        hh_file = Path.joinpath(
            DIR_data,
            "slc_"
            + ID_campaign
            + ID_flight
            + ID_Pass
            + "_"
            + band
            + "hh_t"
            + n_try
            + band
            + ".rat",
        )
        vv_file = Path.joinpath(
            DIR_data,
            "slc_"
            + ID_campaign
            + ID_flight
            + ID_Pass
            + "_"
            + band
            + "vv_t"
            + n_try
            + band
            + ".rat",
        )
        hv_file = Path.joinpath(
            DIR_data,
            "slc_"
            + ID_campaign
            + ID_flight
            + ID_Pass
            + "_"
            + band
            + "hv_t"
            + n_try
            + band
            + ".rat",
        )
        vh_file = Path.joinpath(
            DIR_data,
            "slc_"
            + ID_campaign
            + ID_flight
            + ID_Pass
            + "_"
            + band
            + "vh_t"
            + n_try
            + band
            + ".rat",
        )
        check_files_if_exist(hh_file, vv_file, vh_file, vh_file)
        files = {
            "HH": str(hh_file),
            "VV": str(vv_file),
            "HV": str(hv_file),
            "VH": str(vh_file),
        }
        return files



import numpy as np
from scipy import signal


def calc_hh_vv_ratio_cpd(scat_matrix):
    from scipy import signal

    n_window = 7
    mean_filter = np.ones((n_window, n_window))
    mean_filter /= sum(mean_filter)
    P_hh_filtered = signal.convolve(
        scat_matrix[:, :, 0] * np.conj(scat_matrix[:, :, 0]),
        mean_filter,
        mode="same",
    )
    P_vv_filtered = signal.convolve(
        scat_matrix[:, :, 3] * np.conj(scat_matrix[:, :, 3]),
        mean_filter,
        mode="same",
    )
    p_hh_vv_ratio = (P_hh_filtered / P_vv_filtered).real
    cpd_multilook = signal.convolve(
        (scat_matrix[:, :, 0] * np.conj(scat_matrix[:, :, 3])),
        mean_filter,
        mode="same",
    )
    cpd_deg = (np.angle(cpd_multilook, deg=True)).real

    return p_hh_vv_ratio, cpd_deg



__all__ = ["lex_vec", "pauli_vec"]

import warnings

import numpy as np
from numba import jit, njit
from scipy import signal

# class lex_vec(s_matrix):

#    def __init__(self, *args, **kwargs):
#        self.name = "Lexicographic Scattering Vector"
#        self.allowed_ndim = [4]
#        self.blockprocess = True


def lex_vec(s_matrix, *args, **kwargs):
    """
    𝑘⃗_3𝐿 =[ 𝑆_𝐻𝐻, √2 𝑆_𝐻𝑉, 𝑆_𝑉𝑉 ]𝑇
    """
    if not np.array_equal(s_matrix[:, :, 1], s_matrix[:, :, 2]):
        """
        3.2.2 BISTATIC SCATTERING CASE
        4-D Lexicographic feature vector
        """
        print("4-D Lexicographic feature vector")
        lex_vector = np.zeros(
            (s_matrix.shape[0], s_matrix.shape[1], 4), dtype=complex
        )
        lex_vector[:, :, 0] = s_matrix[:, :, 0]
        lex_vector[:, :, 1] = s_matrix[:, :, 1]
        lex_vector[:, :, 2] = s_matrix[:, :, 2]
        lex_vector[:, :, 3] = s_matrix[:, :, 3]

    else:
        """
        3.2.3 MONOSTATIC BACKSCATTERING CASE
        3-D Lexicographic feature vector
        """
        print("3-D Lexicographic feature vector")
        lex_vector = np.zeros(
            (s_matrix.shape[0], s_matrix.shape[1], 3), dtype=complex
        )
        lex_vector[:, :, 0] = s_matrix[:, :, 0]
        lex_vector[:, :, 1] = np.sqrt(2) * s_matrix[:, :, 1]
        lex_vector[:, :, 2] = s_matrix[:, :, 3]
    return lex_vector


# class pauli_vec(*args, **kwargs):

# def __init__(*args, **kwargs):
#    self.name = "Pauli Scattering Vector"
#    self.allowed_ndim = [4]


def pauli_vec(s_matrix, *args, **kwargs):
    """
    Pauli scattering vector
    𝑘⃗_3𝑃 = 1/√2 [ 𝑆_𝐻𝐻+𝑆_𝑉𝑉, 𝑆_𝐻𝐻−𝑆_𝑉𝑉, 2𝑆_𝐻𝑉 ]𝑇
    """
    if not np.array_equal(s_matrix[:, :, 1], s_matrix[:, :, 2]):
        """
        3.2.2 BISTATIC SCATTERING CASE
        4-D Pauli feature vector
        𝑘⃗_4𝑃 = 
                ⎡   (S₁₁ + S₂₂)   ⎤
                ⎢                 ⎥
        1/√2 .  ⎢   (S₁₁ - S₂₂)   ⎥
                ⎢                 ⎥
                ⎢   (S₁₂ + S₂₁)   ⎥
                ⎢                 ⎥
                ⎣   ⅈ⋅(S₁₂ - S₂₁)  ⎦

        """
        print("4-D Pauli feature vector")
        pauli_vector = np.zeros(
            (s_matrix.shape[0], s_matrix.shape[1], 4), dtype=complex
        )
        pauli_vector[:, :, 0] = s_matrix[:, :, 0] + s_matrix[:, :, 3]
        pauli_vector[:, :, 1] = s_matrix[:, :, 0] - s_matrix[:, :, 3]
        pauli_vector[:, :, 2] = s_matrix[:, :, 1] + s_matrix[:, :, 2]
        pauli_vector[:, :, 3] = 1j * (s_matrix[:, :, 1] - s_matrix[:, :, 2])
        pauli_vector = pauli_vector / np.sqrt(2)
    else:
        """
        3.2.3 MONOSTATIC BACKSCATTERING CASE
        3-D Pauli feature vector
        𝑘⃗_3𝑃 = 
                ⎡   (Sxx + Syy)   ⎤
                ⎢                 ⎥
        1/√2 .  ⎢   (Sxx - Syy)   ⎥
                ⎢                 ⎥
                ⎣   2 . (Sxy )    ⎦
        """
        print("3-D Pauli feature vector")
        pauli_vector = np.zeros(
            (s_matrix.shape[0], s_matrix.shape[1], 3), dtype=complex
        )
        pauli_vector[:, :, 0] = s_matrix[:, :, 0] + s_matrix[:, :, 3]
        pauli_vector[:, :, 1] = s_matrix[:, :, 0] - s_matrix[:, :, 3]
        pauli_vector[:, :, 2] = 2 * s_matrix[:, :, 1]
        pauli_vector = pauli_vector / np.sqrt(2)
    return pauli_vector


def polarimetric_coherency_t_matrix(k_vector, kernel_size=None):
    """
    Calculate the Coherency Matrix [T] and visualize the elements T11, T22, T33 as powers and the elements T13, T23, T12 as powers and their phases. Plus calculate the histograms for everything
    3.5.2 ..................................................... pg 83
    Polarimetric coherency matrix [T 3x3]

    """

    # k_vector = pauli_scattering_vector(S_Matrix)
    pol_coherency_matrix = np.zeros(
        (
            k_vector.shape[0],
            k_vector.shape[1],
            k_vector.shape[2],
            k_vector.shape[2],
        ),
        dtype=complex,
    )

    pol_mat_temp = np.zeros(
        (k_vector.shape[2], k_vector.shape[2]), dtype=complex
    )

    k_vec_temp = np.zeros((k_vector.shape[2], 1), dtype=complex)

    for ix, iy in np.ndindex(k_vector.shape[0:2]):
        k_vec_temp = k_vector[ix, iy, :].reshape(-1, 1)
        pol_mat_temp = np.dot(k_vec_temp, k_vec_temp.T.conjugate())
        pol_coherency_matrix[ix, iy, :, :] = pol_mat_temp

    if kernel_size is None:
        kernel_size = 7
    mean_filter = np.ones((kernel_size, kernel_size))
    mean_filter /= sum(mean_filter)
    pol_coherency_matrix_filtered = np.zeros_like(pol_coherency_matrix)
    for ix, iy in np.ndindex(pol_coherency_matrix.shape[2:]):
        pol_coherency_matrix_filtered[:, :, ix, iy] = signal.convolve(
            pol_coherency_matrix[:, :, ix, iy], mean_filter, mode="same"
        )
    return pol_coherency_matrix, pol_coherency_matrix_filtered


def polarimetric_covariance_c_matrix(omega_vector):
    """
    Calculate the Covariance Matrix [C] and visualize the elements C1, C22, C33 as powers and the elements C13, C23, C12 as powers and their phases. Plus calculate the histograms for everything
    3.5.3 .................................................... 84
    Polarimetric covariance matrix [𝐶 3x3] or [C 4x4]
    """
    # omega_vector = lexicographic_scattering_vector(S_Matrix)
    pol_covariance_matrix = np.zeros(
        (
            omega_vector.shape[0],
            omega_vector.shape[1],
            omega_vector.shape[2],
            omega_vector.shape[2],
        ),
        dtype=complex,
    )

    pol_mat_temp = np.zeros(
        (omega_vector.shape[2], omega_vector.shape[2]), dtype=complex
    )

    omega_vec_temp = np.zeros((omega_vector.shape[2], 1), dtype=complex)
    for ix, iy in np.ndindex(omega_vector.shape[0:2]):
        omega_vec_temp = omega_vector[ix, iy, :].reshape(-1, 1)
        pol_mat_temp = np.dot(omega_vec_temp, omega_vec_temp.T.conjugate())
        pol_covariance_matrix[ix, iy, :, :] = pol_mat_temp

    n_window = 7
    mean_filter = np.ones((n_window, n_window))
    mean_filter /= sum(mean_filter)
    pol_covariance_matrix_filtered = np.zeros_like(pol_covariance_matrix)
    for ix, iy in np.ndindex(pol_covariance_matrix.shape[2:]):
        # print(i)
        pol_covariance_matrix_filtered[:, :, ix, iy] = signal.convolve(
            pol_covariance_matrix[:, :, ix, iy], mean_filter, mode="same"
        )
    return pol_covariance_matrix, pol_covariance_matrix_filtered


def mean_filter(im, mysize=None, noise=None):
    im = np.asarray(im)
    if mysize is None:
        mysize = [3] * im.ndim
    mysize = np.asarray(mysize)
    if mysize.shape == ():
        mysize = np.repeat(mysize.item(), im.ndim)
    mean_filter = np.ones(mysize)
    mean_filter /= sum(mean_filter)

    out = signal.convolve(im, mean_filter, mode="same")
    return out


def dot_prod(k_vector):
    # print(k_vector.shape)
    pol_coherency_matrix = np.zeros(
        (
            k_vector.shape[0],
            k_vector.shape[1],
            k_vector.shape[2],
            k_vector.shape[2],
        ),
        dtype=np.complex_,
    )

    for ix, iy in np.ndindex(k_vector.shape[0:2]):
        # k_vec_temp = k_vector[ix, iy, :]
        # pol_mat_temp = np.dot(k_vector[ix, iy, :], k_vector[ix, iy, :].T.conjugate())
        # pol_coherency_matrix[ix, iy, :, :] = pol_mat_temp
        pol_coherency_matrix[ix, iy, :, :] = np.dot(
            k_vector[ix, iy, :], k_vector[ix, iy, :].T.conjugate()
        )
        # print(ix)

    return pol_coherency_matrix


@jit(nopython=True, nogil=True, cache=True, parallel=False)
def dot_prod_jit(k_vector):
    # print(k_vector.shape)
    pol_coherency_matrix = np.zeros(
        (
            k_vector.shape[0],
            k_vector.shape[1],
            k_vector.shape[2],
            k_vector.shape[2],
        ),
        dtype=np.complex_,
    )

    for ix, iy in np.ndindex(k_vector.shape[0:2]):
        k_vec_temp = k_vector[ix, iy, :].reshape(-1, 1)
        # pol_mat_temp = np.dot(k_vector[ix, iy, :], k_vector[ix, iy, :].T.conjugate())
        # pol_coherency_matrix[ix, iy, :, :] = pol_mat_temp
        # pol_coherency_matrix[ix, iy, :, :] = np.dot(k_vector[ix, iy, :], k_vector[ix, iy, :].T.conjugate())
        # print(ix)
        pol_coherency_matrix[ix, iy, :, :] = np.dot(
            k_vec_temp, k_vec_temp.T.conjugate()
        )

    return pol_coherency_matrix


# @jit(nopython=False, nogil=True, cache=True, parallel=False)
def polarimetric_matrix(k_omege_vector, mysize=None, noise=None):
    """
    from multiprocessing import Pool
    import os
    pool = Pool(os.cpu_count())
    """
    pol_coherency_matrix = dot_prod(k_vector=k_omege_vector)

    im = np.asarray(pol_coherency_matrix)
    if mysize is None:
        mysize = [3] * (im.ndim - 2)
    mysize = np.asarray(mysize)
    if mysize.shape == ():
        mysize = np.repeat(mysize.item(), im.ndim)
    mean_filter = np.ones(mysize, dtype=np.complex_)
    # print(im.shape)
    # pol_coherency_matrix = mean_filter_t_c_matrix(pol_coherency_matrix, mean_filter)
    pol_coherency_matrix_filtered = np.zeros_like(pol_coherency_matrix)
    # print(k_omege_vector.shape)
    for ix, iy in np.ndindex(pol_coherency_matrix.shape[2:]):
        pol_coherency_matrix_filtered[:, :, ix, iy] = signal.convolve(
            pol_coherency_matrix[:, :, ix, iy], mean_filter, mode="same"
        )

    return pol_coherency_matrix / np.sum(mean_filter, axis=None)


# @jit(nopython=False, nogil=True, cache=True, parallel=False)
def polarimetric_matrix_jit(k_omege_vector, mysize=None, noise=None):
    """
    from multiprocessing import Pool
    import os
    pool = Pool(os.cpu_count())
    """
    pol_coherency_matrix = dot_prod_jit(k_vector=k_omege_vector)

    """    
    im = np.asarray(pol_coherency_matrix)

    im_size = pol_coherency_matrix.shape[2:]

    if mysize is None:
        mysize = [3] * (im_size.ndim-2)
    mysize = np.asarray(mysize)
    if mysize.shape == ():
        mysize = np.repeat(mysize.item(), im.ndim)
    """
    mean_filter = np.ones(mysize, dtype=np.complex_)
    n_window = 7
    mean_filter = np.ones((n_window, n_window))

    nx = pol_coherency_matrix.shape[2]
    ny = pol_coherency_matrix.shape[3]
    t_size = pol_coherency_matrix.shape[2:]
    # print(im.shape)
    # pol_coherency_matrix = mean_filter_t_c_matrix(pol_coherency_matrix, mean_filter)
    pol_coherency_matrix_filtered = np.zeros_like(pol_coherency_matrix)
    # print(k_omege_vector.shape)

    il_lower = np.tril_indices(pol_coherency_matrix.shape[3], 0)
    for nx, ny in zip(il_lower[0], il_lower[1]):
        # Multi-look only for lower tri
        # print(nx, ny)
        # print(c_t_matrix[:,:, nx, ny])
        pol_coherency_matrix_filtered[:, :, nx, ny] = signal.convolve(
            pol_coherency_matrix[:, :, nx, ny], mean_filter, mode="same"
        )

    """    
    for i in range(nx):
        for j in range(ny):
            pol_coherency_matrix_filtered[:, :, i, j] = signal.convolve(
            pol_coherency_matrix[:, :, i, j],
            mean_filter, mode="same")
    """
    pol_coherency_matrix_filtered = pol_coherency_matrix_filtered / np.sum(
        mean_filter, axis=None
    )

    return pol_coherency_matrix_filtered


def convolve_t_c_matrix(pol_coherency_matrix):
    pol_coherency_matrix_filtered = np.zeros_like(pol_coherency_matrix)
    il_lower = np.tril_indices(pol_coherency_matrix.shape[3], 0)
    for nx, ny in zip(il_lower[0], il_lower[1]):
        # Multi-look only for lower tri
        # print(nx, ny)
        # print(c_t_matrix[:,:, nx, ny])
        pol_coherency_matrix_filtered[:, :, nx, ny] = signal.convolve(
            pol_coherency_matrix[:, :, nx, ny], mean_filter, mode="same"
        )
    return pol_coherency_matrix_filtered



import sys
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.colors import LogNorm

# # Configure plot universal property
plt.rcParams["figure.figsize"] = (16, 16)
plt.rcParams["font.size"] = 16


def image_dB(y):
    y[y == 0] = sys.float_info.epsilon
    return 10 * np.log10(y)


def stack_rgb(r, g, b):
    """    r_perc = np.percentile(np.abs(r), [0, 100])
    g_perc = np.percentile(np.abs(g), [0, 100])
    b_perc = np.percentile(np.abs(b), [0, 100])
    """
    rgb = np.stack((r, g, b), axis=2)  # stacks 3 h x w arrays -> h x w x 3
    # rgb[ rgb == 0 ] = sys.float_info.epsilon
    # rgb = dB(rgb)

    rgb = image_dB(np.abs(rgb))
    # rgb = np.interp(np.abs(rgb), (np.amin(np.abs(rgb)), 3*np.mean(np.abs(rgb))), (0, 1))
    return rgb


def stack_rgb_linear(r, g, b):
    rgb = np.stack((r, g, b), axis=2)  # stacks 3 h x w arrays -> h x w x 3
    # rgb[ rgb == 0 ] = sys.float_info.epsilon
    rgb = np.interp(
        np.abs(rgb), (np.amin(np.abs(rgb)), 3 * np.mean(np.abs(rgb))), (0, 1)
    )
    return rgb


def plot_pauli_lexi_rbg_hist(
    NAME,
    pauli_vector=None,
    lexicographic_vector=None,
    titles=["Pauli feature vector", "Lexicographic feature vector"],
    plot_type=["rbg"],
):
    Path("./results").mkdir(parents=True, exist_ok=True)

    n_images = 1
    n_plot = len(plot_type)
    print(n_images, n_plot)
    if pauli_vector is not None:
        pauli_vec_rgb = stack_rgb(
            r=pauli_vector[:, :, 1],
            g=pauli_vector[:, :, 2],
            b=pauli_vector[:, :, 0],
        )
        n_images += 1
    if lexicographic_vector is not None:
        lexi_vec_rgb = stack_rgb(
            r=lexicographic_vector[:, :, 0],
            g=lexicographic_vector[:, :, 1],
            b=lexicographic_vector[:, :, 2],
        )
        n_images += 1

    fig = plt.figure(constrained_layout=False, figsize=(24, 12 * n_images))
    gs = fig.add_gridspec(nrows=1, ncols=n_images)

    if "rbg" in plot_type:

        f_ax = fig.add_subplot(gs[0, 0])
        print(
            "Image min:",
            pauli_vec_rgb.min(),
            "Image mean:",
            pauli_vec_rgb.mean(),
            ",Image max:",
            pauli_vec_rgb.max(),
        )
        # pauli_vec_rgb_norm = np.interp(decibel_to_linear(np.abs(pauli_vec_rgb)), (decibel_to_linear(np.amin(np.abs(pauli_vec_rgb))), np.max(decibel_to_linear(np.abs(pauli_vec_rgb)))), (0, 1))
        rgb_pauli = stack_rgb_linear(
            r=pauli_vector[:, :, 1],
            g=pauli_vector[:, :, 2],
            b=pauli_vector[:, :, 0],
        )
        rgb_pauli = np.interp(
            rgb_pauli, (np.amin(rgb_pauli), np.amax(rgb_pauli)), (0, 1)
        )
        ax_ang = f_ax.imshow(
            rgb_pauli,  # vmin=pauli_vec_rgb.min(), vmax=pauli_vec_rgb.max(),
            # aspect=pauli_vec_rgb.shape[1] / pauli_vec_rgb.shape[0] * 2,
            origin="lower",
        )
        f_ax.axis("off")

        # Save just the portion _inside_ the second axis's boundaries
        extent = ax_ang.get_window_extent().transformed(
            fig.dpi_scale_trans.inverted()
        )
        fig.savefig("./results/" + NAME + "_pauli.png", bbox_inches="tight")
        f_ax.set_title(titles[0])

        if lexicographic_vector is not None:
            f_ax = fig.add_subplot(gs[0, 1])
            print(
                "Image min:",
                lexi_vec_rgb.min(),
                "Image mean:",
                lexi_vec_rgb.mean(),
                ",Image max:",
                lexi_vec_rgb.max(),
            )
            lexi_vec_rgb_norm = np.interp(
                np.abs(lexi_vec_rgb),
                (np.amin(np.abs(lexi_vec_rgb)), np.max(np.abs(lexi_vec_rgb))),
                (0, 1),
            )

            rgb_lex = stack_rgb_linear(
                r=lexicographic_vector[:, :, 0],
                g=lexicographic_vector[:, :, 1],
                b=lexicographic_vector[:, :, 2],
            )
            ax_ang = f_ax.imshow(
                rgb_lex,  # vmin=10, vmax=200, # vmin=lexi_vec_rgb.min(), vmax=lexi_vec_rgb.max(),
                # aspect=lexi_vec_rgb.shape[1] / lexi_vec_rgb.shape[0] * 2,
                origin="lower",
            )
            f_ax.set_title(titles[1])
            f_ax.axis("off")

    if "hist" in plot_type:
        num_bins = 10000

        f_ax2 = fig.add_subplot(gs[0, -1])

        color = ("r", "g", "b")
        histogram = ["$S_{XX} - S_{YY}$", "$S_{XY}$", "$S_{XX} + S_{YY}$"]
        for i, col in enumerate(color):
            chan = pauli_vec_rgb[:, :, i].ravel()
            # f_ax2.hist(chan, bins='auto', color = col, histtype='step', label=histogram[i], density=True, linestyle=('solid'))
            #    f_cc = fig.add_subplot(gs[0, 1])
            g = sns.kdeplot(
                (chan.flatten()),
                shade=True,
                label=histogram[i],
                color=col,
                ax=f_ax2,
            )
        g.legend()
        g.legend_.set_title("Pauli")

        if lexicographic_vector != None:
            histogram = ["$S_{XX}$", r"$\sqrt{2} S_{XY}$", "$S_{YY}$"]
            for i, col in enumerate(color):
                chan = lexi_vec_rgb[:, :, i].ravel()
                f_ax2.hist(
                    chan,
                    bins="auto",
                    color=col,
                    histtype="step",
                    label=histogram[i],
                    density=True,
                    linestyle=(":"),
                )

        f_ax2.set_xlabel("Value (dB)")
        f_ax2.set_ylabel("Frequency")
        im_perc = np.percentile((pauli_vec_rgb), [0.01, 99.99])
        f_ax2.set_xlim(im_perc[0], im_perc[1])
        f_ax2.set_title("Histogram")

    fig.suptitle(NAME)  # or plt.suptitle('Main title')
    plt.savefig("./results/" + NAME + "_pauli_full.png")

    # sns.plt.show()

    return fig


def bytescale(data, cmin=None, cmax=None, high=255, low=0):
    """
    Byte scales an array (image). Byte scaling means converting the input image to uint8 dtype and scaling
    the range to ``(low, high)`` (default 0-255).     If the input image already has dtype uint8, no scaling
    is done.

    :author: taken from scipy.misc (deprecated in scipy, will disappear in version 1.2)

    :param data: image data array
    :type img1: ndarray
    :param cmin: Bias scaling of small values. Default is data.min()
    :type cmin: scalar, optional
    :param cmax: Bias scaling of large values. Default is data.max()
    :type cmax: scalar, optional
    :param high: Scale max value to high. Default is 255.
    :type high: scalar, optional
    :param low: Scale min value to low. Default is 0.
    :type low: scalar, optional
    :returns: The byte-scaled array as ndarray
    """

    if data.dtype == np.uint8:
        return data

    if high < low:
        raise ValueError("`high` should be larger than `low`.")

    if cmin is None:
        cmin = data.min()
    if cmax is None:
        cmax = data.max()

    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError("`cmax` should be larger than `cmin`.")
    elif cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data * 1.0 - cmin) * scale + 0.4999
    bytedata[bytedata > high] = high
    bytedata[bytedata < 0] = 0
    return np.cast[np.uint8](bytedata) + np.cast[np.uint8](low)


def plot_pauli_lexi_rbg_hist_bytescale(
    NAME,
    pauli_vector=None,
    lexicographic_vector=None,
    im_bytescale=False,
    titles=["Pauli feature vector", "Lexicographic feature vector"],
    plot_type=["rbg"],
    *args,
    **kwargs,
):
    n_images = len(titles)
    n_plot = len(plot_type)

    fig = plt.figure(constrained_layout=False, figsize=(12, 24))
    gs = fig.add_gridspec(nrows=n_plot, ncols=n_images)

    if "rbg" in plot_type:
        if pauli_vector is not None:
            f_ax = fig.add_subplot(gs[0, 0])
            pauli_vec_rgb = stack_rgb(
                r=pauli_vector[:, :, 1],
                g=pauli_vector[:, :, 2],
                b=pauli_vector[:, :, 0],
            )
            print(
                "Image min:",
                pauli_vec_rgb.min(),
                "Image mean:",
                pauli_vec_rgb.mean(),
                ",Image max:",
                pauli_vec_rgb.max(),
            )
            # pauli_vec_rgb_norm = np.interp(decibel_to_linear(np.abs(pauli_vec_rgb)), (decibel_to_linear(np.amin(np.abs(pauli_vec_rgb))), np.max(decibel_to_linear(np.abs(pauli_vec_rgb)))), (0, 1))
            if im_bytescale is True:
                k_vec = np.abs(pauli_vector)
                k_sca = k_vec ** 0.7
                """
                vmin=2
                vmax=98
                im_perc = np.percentile(np.abs(k_sca), [vmin, vmax])
                """
                k_1 = bytescale(
                    k_sca[:, :, 0], cmin=k_sca.min(), cmax=k_sca.mean()
                )
                k_2 = bytescale(
                    k_sca[:, :, 1], cmin=k_sca.min(), cmax=k_sca.mean()
                )
                k_3 = bytescale(
                    k_sca[:, :, 2], cmin=k_sca.min(), cmax=k_sca.mean()
                )
                rgb = np.stack((k_2, k_3, k_1), axis=2)
                ax_ang = f_ax.imshow(
                    rgb,  # vmin=pauli_vec_rgb.min(), vmax=pauli_vec_rgb.max(),
                    *args,
                    **kwargs,
                )
            else:
                ax_ang = f_ax.imshow(
                    stack_rgb_linear(
                        r=pauli_vector[:, :, 1],
                        g=pauli_vector[:, :, 2],
                        b=pauli_vector[:, :, 0],
                    ),  # vmin=pauli_vec_rgb.min(), vmax=pauli_vec_rgb.max(),
                    *args,
                    **kwargs,
                )
            f_ax.set_title(titles[0])
            f_ax.axis("off")

        if lexicographic_vector is not None:
            lexi_vec_rgb = stack_rgb(
                r=lexicographic_vector[:, :, 0],
                g=lexicographic_vector[:, :, 1],
                b=lexicographic_vector[:, :, 2],
            )
            f_ax = fig.add_subplot(gs[0, 1])
            print(
                "Image min:",
                lexi_vec_rgb.min(),
                "Image mean:",
                lexi_vec_rgb.mean(),
                ",Image max:",
                lexi_vec_rgb.max(),
            )
            lexi_vec_rgb_norm = np.interp(
                np.abs(lexi_vec_rgb),
                (np.amin(np.abs(lexi_vec_rgb)), np.max(np.abs(lexi_vec_rgb))),
                (0, 1),
            )
            if im_bytescale is True:
                k_vec = np.abs(lexicographic_vector)
                k_sca = k_vec ** 0.7
                vmin = 0
                vmax = 98
                im_perc = np.percentile(np.abs(k_sca), [vmin, vmax])
                k_1 = bytescale(
                    abs(k_sca[:, :, 0]), cmin=im_perc[0], cmax=1 * k_sca.mean()
                )
                k_2 = bytescale(
                    abs(k_sca[:, :, 1]), cmin=im_perc[0], cmax=1 * k_sca.mean()
                )
                k_3 = bytescale(
                    abs(k_sca[:, :, 2]), cmin=im_perc[0], cmax=1 * k_sca.mean()
                )
                rgb = np.stack((k_1, k_2, k_3), axis=2)
                ax_ang = f_ax.imshow(rgb, *args, **kwargs)

            else:
                ax_ang = f_ax.imshow(
                    stack_rgb_linear(
                        r=lexicographic_vector[:, :, 0],
                        g=lexicographic_vector[:, :, 1],
                        b=lexicographic_vector[:, :, 2],
                    ),  # vmin=10, vmax=200, # vmin=lexi_vec_rgb.min(), vmax=lexi_vec_rgb.max(),
                    *args,
                    **kwargs,
                )
            f_ax.set_title(titles[-1])
            f_ax.axis("off")

    if "hist" in plot_type:
        num_bins = 10000
        colors = ["#E8000B", "#03ED3A", "#003FFF", "#1A1A1A"]  # R, G, B,Black
        palette = sns.color_palette(colors, 4)
        f_ax2 = fig.add_subplot(gs[1, :])

        if pauli_vector is not None:

            histogram = [
                "${S_{HH} - S_{VV}}_{(Dihedral)}$",
                "${2 S_{HV}}_{(Volume)}$",
                "${S_{HH} + S_{VV} }_{(Surface)}$",
            ]
            for i, col in zip(range(pauli_vec_rgb.shape[2]), palette):
                chan = pauli_vec_rgb[:, :, i].ravel()
                f_ax2.hist(
                    chan,
                    bins="auto",
                    color=col,
                    histtype="step",
                    label=histogram[i],
                    density=True,
                    linestyle=("solid"),
                )

        if lexicographic_vector is not None:
            histogram = ["$S_{HH}$", "$\sqrt{2} S_{HV}$", "$S_{VV}$"]
            for i, col in zip(range(lexi_vec_rgb.shape[2]), palette):
                chan = lexi_vec_rgb[:, :, i].ravel()
                f_ax2.hist(
                    chan,
                    bins="auto",
                    color=col,
                    histtype="step",
                    label=histogram[i],
                    density=True,
                    linestyle=(":"),
                )

        f_ax2.set_xlabel("Value")
        f_ax2.set_ylabel("Frequency")
        im_perc = np.percentile((pauli_vec_rgb), [0.01, 99.99])
        f_ax2.set_xlim(im_perc[0], im_perc[1])
        f_ax2.set_title("Histogram")

        plt.legend(loc="upper left")

    return fig



import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.colors import LogNorm
from matplotlib.ticker import PercentFormatter

from ..utils.plot import stack_rgb, stack_rgb_linear


def bytescale(data, cmin=None, cmax=None, high=255, low=0):
    """
    Byte scales an array (image). Byte scaling means converting the input image to uint8 dtype and scaling
    the range to ``(low, high)`` (default 0-255).     If the input image already has dtype uint8, no scaling
    is done.

    :author: taken from scipy.misc (deprecated in scipy, will disappear in version 1.2)

    :param data: image data array
    :type img1: ndarray
    :param cmin: Bias scaling of small values. Default is data.min()
    :type cmin: scalar, optional
    :param cmax: Bias scaling of large values. Default is data.max()
    :type cmax: scalar, optional
    :param high: Scale max value to high. Default is 255.
    :type high: scalar, optional
    :param low: Scale min value to low. Default is 0.
    :type low: scalar, optional
    :returns: The byte-scaled array as ndarray
    """

    if data.dtype == np.uint8:
        return data

    if high < low:
        raise ValueError("`high` should be larger than `low`.")

    if cmin is None:
        cmin = data.min()
    if cmax is None:
        cmax = data.max()

    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError("`cmax` should be larger than `cmin`.")
    elif cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data * 1.0 - cmin) * scale + 0.4999
    bytedata[bytedata > high] = high
    bytedata[bytedata < 0] = 0
    return np.cast[np.uint8](bytedata) + np.cast[np.uint8](low)


def plot_pauli_entropy_ani_alpha_vv_hh_phi_diff(
    t_matrix,
    slc_ent_ani_alp_44_dic,
    p_hh_vv_ratio=None,
    cpd_deg=None,
    aspect=None,
    suptitle="Pauli, Entropy, Ani, Alpha, HH-VV Ratio, CPD",
    im_bytescale=True,
    *args,
    **kwargs,
):
    n_images = 6  # len(titles)
    n_plot = 1  # len(plot_type)

    # fig = plt.figure(constrained_layout=False, figsize=(5*n_images, 20))
    fig, axs = plt.subplots(
        1, 6, constrained_layout=True, figsize=(5 * n_images, 10)
    )
    # gs = fig.add_gridspec(nrows=n_plot, ncols=n_images)

    # Pauli
    pauli_vector = np.stack(
        (t_matrix[:, :, 0, 0], t_matrix[:, :, 1, 1], t_matrix[:, :, 2, 2]),
        axis=2,
    )
    if aspect is None:
        aspect = (pauli_vector.shape[1] / pauli_vector.shape[0]) / (1785 / 3140)
    elif aspect is None:
        aspect = (pauli_vector.shape[1] / pauli_vector.shape[0]) / aspect
    # f_ax = fig.add_subplot(gs[0, 0])
    f_ax = axs[0]
    pauli_vec_rgb = stack_rgb(
        r=pauli_vector[:, :, 1],
        g=pauli_vector[:, :, 2],
        b=pauli_vector[:, :, 0],
    )
    print(
        "Image min:",
        pauli_vec_rgb.min(),
        "Image mean:",
        pauli_vec_rgb.mean(),
        ",Image max:",
        pauli_vec_rgb.max(),
    )
    # pauli_vec_rgb_norm = np.interp(decibel_to_linear(np.abs(pauli_vec_rgb)), (decibel_to_linear(np.amin(np.abs(pauli_vec_rgb))), np.max(decibel_to_linear(np.abs(pauli_vec_rgb)))), (0, 1))
    if im_bytescale is True:
        k_vec = np.abs(pauli_vector)
        k_sca = k_vec ** 0.7
        """
        vmin=2
        vmax=98
        im_perc = np.percentile(np.abs(k_sca), [vmin, vmax])
        """
        k_1 = bytescale(k_sca[:, :, 0], cmin=k_sca.min(), cmax=k_sca.mean())
        k_2 = bytescale(k_sca[:, :, 1], cmin=k_sca.min(), cmax=k_sca.mean())
        k_3 = bytescale(k_sca[:, :, 2], cmin=k_sca.min(), cmax=k_sca.mean())
        rgb = np.stack((k_2, k_3, k_1), axis=2)
        ax_ang = f_ax.imshow(
            rgb,  # vmin=pauli_vec_rgb.min(), vmax=pauli_vec_rgb.max(),
            aspect=aspect,
        )
    else:
        ax_ang = f_ax.imshow(
            stack_rgb_linear(
                r=pauli_vector[:, :, 1],
                g=pauli_vector[:, :, 2],
                b=pauli_vector[:, :, 0],
            ),  # vmin=pauli_vec_rgb.min(), vmax=pauli_vec_rgb.max(),
            aspect=aspect,
            *args,
            **kwargs,
        )
    f_ax.set_title("Pauli", fontsize=24)
    f_ax.axis("off")
    f_ax.set_anchor(anchor="N", share=True)

    for idx, key in enumerate(slc_ent_ani_alp_44_dic):
        title = key
        array = slc_ent_ani_alp_44_dic[key]
        ax_im = axs[idx + 1]
        # ax_im = fig.add_subplot(gs[:, idx+1])
        """        
        if (np.max(array) - np.min(array)) < 1.1:
            ax_cc = ax_im.imshow(array,  cmap='jet',
                                vmin=0.0, vmax=1.)
        elif (np.max(array) - np.min(array)) < 100 and (np.max(array) - np.min(array)) > 10:
            ax_cc = ax_im.imshow(array, cmap='jet',
                                vmin=0.0, vmax=90.)
        else:
        """
        ax_cc = ax_im.imshow(
            array,
            cmap="jet",
            aspect=aspect,
            vmin=np.min(array),
            vmax=np.max(array),
        )

        ax_im.axis("off")
        ax_im.set_title(title, fontsize=24)
        # ax_im.clim(0, 10)
        if ax_cc.get_clim()[1] <= 1:

            ax_cc.set_clim(0, 1)
        elif ax_cc.get_clim()[1] >= 5:
            ax_cc.set_clim(0, 90)
        else:
            ax_cc.set_clim(ax_cc.get_clim()[0], ax_cc.get_clim()[1])

        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_cc,
            ax=ax_im,  # ticks=[-np.pi*.99, 0, np.pi*.99],
            shrink=1,
            location="bottom",
        )
        cbar.ax.set_xlabel("", rotation=0)
        ax_im.set_anchor(anchor="N", share=True)
    cbar.ax.set_xlabel("$[deg]$", rotation=0)
    # cbar.ax.set_title ("$[deg]$")
    # fig.colorbar(ax_cc, ax=axs[1:3], shrink=0.6, location='bottom')
    # fig.colorbar(ax_cc, ax=axs[3], shrink=0.6, location='bottom')

    if p_hh_vv_ratio is not None:
        # ax_im = fig.add_subplot(gs[:, -2])
        ax_im = axs[-2]
        if aspect is None:
            aspect = (p_hh_vv_ratio.shape[1] / p_hh_vv_ratio.shape[0]) / (
                1785 / 3140
            )
        elif aspect is None:
            aspect = (p_hh_vv_ratio.shape[1] / p_hh_vv_ratio.shape[0]) / aspect
        im_perc = np.percentile(p_hh_vv_ratio, [2, 98])

        ax_cc = ax_im.imshow(
            p_hh_vv_ratio,
            cmap="jet",
            aspect=aspect,
            norm=LogNorm(vmin=im_perc[0], vmax=im_perc[1]),
        )
        ax_im.axis("off")
        ax_im.set_title("HH-VV Power Ratio", fontsize=24)
        cbar = fig.colorbar(
            ax_cc,
            ax=ax_im,  # ticks=[-np.pi*.99, 0, np.pi*.99],
            shrink=1,
            location="bottom",
        )
        cbar.ax.set_xlabel("$[dB]$", rotation=0)
        # cbar.ax.set_title ("$[dB]$")
        ax_im.set_anchor(anchor="N", share=True)
        # fig.colorbar(ax_cc, ax=axs[-2], shrink=0.6, location='bottom')

    if cpd_deg is not None:
        # ax_im = fig.add_subplot(gs[:, -1])
        ax_im = axs[-1]
        if aspect is None:
            aspect = (cpd_deg.shape[1] / cpd_deg.shape[0]) / (1785 / 3140)
        elif aspect is None:
            aspect = (cpd_deg.shape[1] / cpd_deg.shape[0]) / aspect

        im_perc = np.percentile(cpd_deg, [5, 95])

        ax_cc = ax_im.imshow(
            cpd_deg, cmap="jet", aspect=aspect, vmin=im_perc[0], vmax=im_perc[1]
        )
        ax_im.axis("off")
        ax_im.set_title("$\phi_{HH-VV}$", fontsize=24)
        # fig.colorbar(ax_cc, ax=axs[-1], shrink=0.6, location='bottom')
        cbar = fig.colorbar(
            ax_cc,
            ax=ax_im,  # ticks=[-np.pi*.99, 0, np.pi*.99],
            shrink=1,
            location="bottom",
        )
        cbar.ax.set_xlabel("$[deg]$", rotation=0)
        # cbar.ax.set_title ("$[deg]$")
        ax_im.set_anchor(anchor="N", share=True)
    fig.suptitle(suptitle)
    return fig


def plot_eigenvalues_hist_dB(
    array_dict,
    suptitle,
    vmin=2,
    vmax=98,
    aspect=None,
    origin="lower",
    *args,
    **kwargs,
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)
    # sns.palplot(palette)

    n_images = len(array_dict.keys())
    fig = plt.figure(constrained_layout=False, figsize=(4 * n_images, 10))
    gs = fig.add_gridspec(nrows=2, ncols=n_images)
    f_img = {}

    ax_img = {}
    f_hist = fig.add_subplot(gs[1, :])

    for idx, key in enumerate(array_dict):
        legend = key
        array = array_dict[key]
        if aspect is None:
            aspect = array.shape[1] / array.shape[0] * 2

        f_img[idx] = fig.add_subplot(gs[0, idx])
        im_perc = np.percentile(array, [vmin, vmax])

        ax_img[idx] = f_img[idx].imshow(
            array,
            aspect=aspect,
            cmap="jet",
            vmin=0,
            vmax=im_perc[1],
            origin=origin,
        )
        f_img[idx].axis("off")
        f_img[idx].set_title(legend)
        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_img[idx],
            ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=0.8,
            pad=0.1,
        )
        cbar.ax.set_xlabel("", rotation=0)

        f_hist.hist(
            image_dB(array.flatten()),
            color=palette[idx],
            bins="auto",
            histtype="step",
            label=legend,
            density=True,
        )  # , linestyle=(':')
        f_hist.legend()
        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))
    f_hist.set_xlabel("$dB$")
    f_hist.set_ylabel("Normalized Frequency")
    fig.suptitle(suptitle)
    # now = datetime.datetime.now()
    # fig.savefig("eigenvalues_{}_{}.tiff".format(title, now.strftime("%Y-%m-%d-%H:%M:%S")))
    return fig


def plot_eigenvalues_hist_dB_2(
    array_dict,
    suptitle,
    array_dict2=None,
    vmin=2,
    vmax=98,
    aspect=None,
    origin="lower",
    *args,
    **kwargs,
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)
    # sns.palplot(palette)

    n_images = len(array_dict.keys())
    fig = plt.figure(constrained_layout=False, figsize=(5 * n_images, 20))
    gs = fig.add_gridspec(nrows=3, ncols=n_images)
    f_img = {}

    ax_img = {}
    f_hist = fig.add_subplot(gs[-1, :])

    for idx, key in enumerate(array_dict):
        legend = key
        array = array_dict[key]
        if aspect is None:
            aspect = array.shape[1] / array.shape[0] * 2

        f_img[idx] = fig.add_subplot(gs[0, idx])
        im_perc = np.percentile(array, [vmin, vmax])

        ax_img[idx] = f_img[idx].imshow(
            array,
            aspect=aspect,
            cmap="jet",
            vmin=0,
            vmax=im_perc[1],
            origin=origin,
        )
        f_img[idx].axis("off")
        f_img[idx].set_title(legend)
        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_img[idx],
            ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=0.8,
            pad=0.1,
        )
        cbar.ax.set_xlabel("", rotation=0)

        f_hist.hist(
            image_dB(array.flatten()),
            color=palette[idx],
            bins="auto",
            histtype="step",
            label=legend,
            density=True,
        )  # , linestyle=(':')
        f_hist.legend()
        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))

    if array_dict2 is not None:
        for idx, key in enumerate(array_dict2):
            legend = key
            array = array_dict2[key]
            if aspect is None:
                aspect = array.shape[1] / array.shape[0] * 2

            f_img[idx] = fig.add_subplot(gs[-2, idx])
            im_perc = np.percentile(array, [vmin, vmax])

            ax_img[idx] = f_img[idx].imshow(
                array,
                aspect=aspect,
                cmap="jet",
                vmin=0,
                vmax=im_perc[1],
                origin=origin,
            )
            f_img[idx].axis("off")
            f_img[idx].set_title(legend)
            # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
            cbar = fig.colorbar(
                ax_img[idx],
                ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
                orientation="vertical",
                shrink=0.8,
                pad=0.1,
            )
            cbar.ax.set_xlabel("", rotation=0)

            f_hist.hist(
                image_dB(array.flatten()),
                color=palette[idx],
                bins="auto",
                histtype="step",
                label=legend,
                density=True,
                linestyle=(":"),
            )  #
            f_hist.legend()

    f_hist.set_xlabel("$dB$")
    f_hist.set_ylabel("Normalized Frequency")
    fig.suptitle(suptitle)
    # now = datetime.datetime.now()
    # fig.savefig("eigenvalues_{}_{}.tiff".format(title, now.strftime("%Y-%m-%d-%H:%M:%S")))
    return fig


def plot_image_hist(
    array_dict, suptitle, aspect=None, origin="lower", *args, **kwargs
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)

    n_images = len(array_dict.keys())
    fig = plt.figure(constrained_layout=False, figsize=(9 * n_images, 20))
    gs = fig.add_gridspec(nrows=2, ncols=n_images + 1)

    for idx, key in enumerate(array_dict):
        title = key
        array = array_dict[key]
        ax_im = fig.add_subplot(gs[0, idx])
        """        
        if (np.max(array) - np.min(array)) < 1.1:
            ax_cc = ax_im.imshow(array,  cmap='jet',
                                vmin=0.0, vmax=1.)
        elif (np.max(array) - np.min(array)) < 100 and (np.max(array) - np.min(array)) > 10:
            ax_cc = ax_im.imshow(array, cmap='jet',
                                vmin=0.0, vmax=90.)
        else:
        """
        ax_cc = ax_im.imshow(
            array, cmap="jet", vmin=np.min(array), vmax=np.max(array)
        )

        ax_im.axis("off")
        ax_im.set_title(title)
        # ax_im.clim(0, 10)
        ax_cc.set_clim(ax_cc.get_clim()[0], ax_cc.get_clim()[1])

        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_cc,
            ax=ax_im,  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=1.0,
            pad=0.05,
        )
        cbar.ax.set_xlabel("", rotation=0)

        ax_hist = fig.add_subplot(gs[1, idx])
        ax_hist.hist(
            (array.ravel()),
            color=palette[idx],
            bins=1000,
            histtype="step",
            label=title,
            density=False,
            weights=100 * np.ones(len(array.ravel())) / len(array.ravel()),
        )  # , linestyle=(':')
        ax_hist.set_xlim(ax_cc.get_clim())
        # Now we format the y-axis to display percentage
        ax_hist.yaxis.set_major_formatter(PercentFormatter(xmax=1))
        fig.suptitle(suptitle)

        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))

    return fig


def plot_entropy_an_alpha_hist(
    array_dict, suptitle, aspect=None, origin="lower", *args, **kwargs
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)

    n_images = len(array_dict.keys())
    fig = plt.figure(constrained_layout=False, figsize=(9 * n_images, 20))
    gs = fig.add_gridspec(nrows=3, ncols=n_images)

    for idx, key in enumerate(array_dict):
        title = key
        array = array_dict[key]
        if aspect is None:
            aspect = (array.shape[1] / array.shape[0]) / (1785 / 3140)
        elif aspect is None:
            aspect = (array.shape[1] / array.shape[0]) / aspect

        ax_im = fig.add_subplot(gs[:2, idx])
        """        
        if (np.max(array) - np.min(array)) < 1.1:
            ax_cc = ax_im.imshow(array,  cmap='jet',
                                vmin=0.0, vmax=1.)
        elif (np.max(array) - np.min(array)) < 100 and (np.max(array) - np.min(array)) > 10:
            ax_cc = ax_im.imshow(array, cmap='jet',
                                vmin=0.0, vmax=90.)
        else:
        """
        ax_cc = ax_im.imshow(
            array,
            cmap="jet",
            aspect=aspect,
            vmin=np.min(array),
            vmax=np.max(array),
        )

        ax_im.axis("off")
        ax_im.set_title(title)
        # ax_im.clim(0, 10)
        if ax_cc.get_clim()[1] <= 1:

            ax_cc.set_clim(0, 1)
        elif ax_cc.get_clim()[1] >= 5:
            ax_cc.set_clim(0, 90)
        else:
            ax_cc.set_clim(ax_cc.get_clim()[0], ax_cc.get_clim()[1])

        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_cc,
            ax=ax_im,  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=0.7,
        )
        cbar.ax.set_xlabel("", rotation=0)

        ax_hist = fig.add_subplot(gs[2, idx])
        ax_hist.hist(
            (array.ravel()),
            color=palette[idx],
            bins=1000,
            histtype="step",
            label=title,
            density=False,
            weights=100 * np.ones(len(array.ravel())) / len(array.ravel()),
        )  # , linestyle=(':')
        ax_hist.set_xlim(ax_cc.get_clim())
        # Now we format the y-axis to display percentage
        ax_hist.yaxis.set_major_formatter(PercentFormatter(xmax=1))
        fig.suptitle(suptitle)

        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))

    return fig


def plot_eigenvalues_entropy_dB(
    array_dict,
    suptitle,
    array_dict2=None,
    array_entropy=None,
    vmin=2,
    vmax=98,
    aspect=None,
    origin="lower",
    *args,
    **kwargs,
):
    colors = [
        "#003FFF",
        "#E8000B",
        "#03ED3A",
        "#1A1A1A",
    ]  # Blue, Red, Green, Black
    palette = sns.color_palette(colors, 4)
    # sns.palplot(palette)

    n_images = len(array_dict.keys()) + 1
    fig = plt.figure(constrained_layout=False, figsize=(5 * n_images, 20))
    gs = fig.add_gridspec(nrows=3, ncols=n_images)
    f_img = {}

    ax_img = {}
    f_hist = fig.add_subplot(gs[-1, 0:4])

    for idx, key in enumerate(array_dict):
        legend = key
        array = array_dict[key]

        if aspect is None:
            aspect = (array.shape[1] / array.shape[0]) / (1785 / 3140)
        elif aspect is None:
            aspect = (array.shape[1] / array.shape[0]) / aspect

        f_img[idx] = fig.add_subplot(gs[0, idx])
        im_perc = np.percentile(array, [vmin, vmax])

        ax_img[idx] = f_img[idx].imshow(
            array,
            aspect=aspect,
            cmap="jet",
            vmin=0,
            vmax=im_perc[1],
            origin=origin,
        )
        f_img[idx].axis("off")
        f_img[idx].set_title(legend)
        # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
        cbar = fig.colorbar(
            ax_img[idx],
            ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
            orientation="vertical",
            shrink=0.8,
            pad=0.1,
        )
        cbar.ax.set_xlabel("", rotation=0)

        f_hist.hist(
            image_dB(array.flatten()),
            color=palette[idx],
            bins="auto",
            histtype="step",
            label=legend,
            density=True,
        )  # , linestyle=(':')
        f_hist.legend()
        # ax = (sns.kdeplot((array.flatten()), shade=False, ax=f_cc, legend=True, label=title))

    if array_dict2 is not None:
        for idx, key in enumerate(array_dict2):
            legend = key
            array = array_dict2[key]
            if aspect is None:
                aspect = (array.shape[1] / array.shape[0]) / (1785 / 3140)
            elif aspect is None:
                aspect = (array.shape[1] / array.shape[0]) / aspect

            f_img[idx] = fig.add_subplot(gs[-2, idx])
            im_perc = np.percentile(array, [vmin, vmax])

            ax_img[idx] = f_img[idx].imshow(
                array,
                aspect=aspect,
                cmap="jet",
                vmin=0,
                vmax=im_perc[1],
                origin=origin,
            )
            f_img[idx].axis("off")
            f_img[idx].set_title(legend)
            # fig.colorbar(ax_ang, ax=f_ax2, shrink=0.6)
            cbar = fig.colorbar(
                ax_img[idx],
                ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
                orientation="vertical",
                shrink=0.8,
                pad=0.1,
            )
            cbar.ax.set_xlabel("", rotation=0)

            f_hist.hist(
                image_dB(array.flatten()),
                color=palette[idx],
                bins="auto",
                histtype="step",
                label=legend,
                density=True,
                linestyle=(":"),
            )  #
            f_hist.legend()
    if array_entropy is not None:
        f_hist_entropy = fig.add_subplot(gs[-1, -1])
        for idx, key in enumerate(array_entropy):
            legend = key
            array = array_entropy[key]
            if aspect is None:
                aspect = (array.shape[1] / array.shape[0]) / (1785 / 3140)
            elif aspect is None:
                aspect = (array.shape[1] / array.shape[0]) / aspect

            f_img[idx] = fig.add_subplot(gs[idx, -1])

            ax_img[idx] = f_img[idx].imshow(
                array, aspect=aspect, cmap="jet", vmin=0, vmax=1, origin=origin
            )
            f_img[idx].axis("off")
            f_img[idx].set_title(legend)

            f_hist_entropy.hist(
                (array.flatten()),
                color=palette[idx],
                bins="auto",
                histtype="step",
                label=legend,
                density=True,
                linestyle=(":"),
            )  #
            f_hist_entropy.legend()
            cbar = fig.colorbar(
                ax_img[idx],
                ax=f_img[idx],  # ticks=[-np.pi*.99, 0, np.pi*.99],
                orientation="vertical",
                shrink=0.8,
                pad=0.1,
            )
            cbar.ax.set_xlabel("", rotation=0)

    f_hist.set_xlabel("$dB$")
    f_hist.set_ylabel("Normalized Frequency")
    fig.suptitle(suptitle)
    # now = datetime.datetime.now()
    # fig.savefig("eigenvalues_{}_{}.tiff".format(title, now.strftime("%Y-%m-%d-%H:%M:%S")))
    return fig


def image_dB(y):
    return 10 * np.log10(y)
